from functools import lru_cache
from typing import Literal

import mlx.core as mx

from mflux.models.common.resolution.config_resolution import ConfigResolution


class ModelConfig:
    precision: mx.Dtype = mx.bfloat16

    def __init__(
        self,
        priority: int,
        aliases: list[str],
        model_name: str,
        base_model: str | None,
        controlnet_model: str | None,
        custom_transformer_model: str | None,
        num_train_steps: int | None,
        max_sequence_length: int | None,
        supports_guidance: bool | None,
        requires_sigma_shift: bool | None,
    ):
        self.aliases = aliases
        self.model_name = model_name
        self.base_model = base_model
        self.controlnet_model = controlnet_model
        self.custom_transformer_model = custom_transformer_model
        self.num_train_steps = num_train_steps
        self.max_sequence_length = max_sequence_length
        self.supports_guidance = supports_guidance
        self.requires_sigma_shift = requires_sigma_shift
        self.priority = priority

    @staticmethod
    @lru_cache
    def dev() -> "ModelConfig":
        return AVAILABLE_MODELS["dev"]

    @staticmethod
    @lru_cache
    def schnell() -> "ModelConfig":
        return AVAILABLE_MODELS["schnell"]

    @staticmethod
    @lru_cache
    def dev_kontext() -> "ModelConfig":
        return AVAILABLE_MODELS["dev-kontext"]

    @staticmethod
    @lru_cache
    def dev_fill() -> "ModelConfig":
        return AVAILABLE_MODELS["dev-fill"]

    @staticmethod
    @lru_cache
    def dev_redux() -> "ModelConfig":
        return AVAILABLE_MODELS["dev-redux"]

    @staticmethod
    @lru_cache
    def dev_depth() -> "ModelConfig":
        return AVAILABLE_MODELS["dev-depth"]

    @staticmethod
    @lru_cache
    def dev_controlnet_canny() -> "ModelConfig":
        return AVAILABLE_MODELS["dev-controlnet-canny"]

    @staticmethod
    @lru_cache
    def schnell_controlnet_canny() -> "ModelConfig":
        return AVAILABLE_MODELS["schnell-controlnet-canny"]

    @staticmethod
    @lru_cache
    def dev_controlnet_upscaler() -> "ModelConfig":
        return AVAILABLE_MODELS["dev-controlnet-upscaler"]

    @staticmethod
    @lru_cache
    def dev_fill_catvton() -> "ModelConfig":
        return AVAILABLE_MODELS["dev-fill-catvton"]

    @staticmethod
    @lru_cache
    def krea_dev() -> "ModelConfig":
        return AVAILABLE_MODELS["krea-dev"]

    @staticmethod
    @lru_cache
    def qwen_image() -> "ModelConfig":
        return AVAILABLE_MODELS["qwen-image"]

    @staticmethod
    @lru_cache
    def qwen_image_edit() -> "ModelConfig":
        return AVAILABLE_MODELS["qwen-image-edit"]

    @staticmethod
    @lru_cache
    def fibo() -> "ModelConfig":
        return AVAILABLE_MODELS["fibo"]

    @staticmethod
    @lru_cache
    def z_image_turbo() -> "ModelConfig":
        return AVAILABLE_MODELS["z-image-turbo"]

    @staticmethod
    @lru_cache
    def seedvr2_3b() -> "ModelConfig":
        return AVAILABLE_MODELS["seedvr2-3b"]

    def x_embedder_input_dim(self) -> int:
        if "Fill" in self.model_name:
            return 384
        if "Depth" in self.model_name:
            return 128
        else:
            return 64

    def is_canny(self) -> bool:
        return self.controlnet_model is not None and "Canny" in self.controlnet_model

    @staticmethod
    def from_name(
        model_name: str,
        base_model: Literal["dev", "schnell", "krea-dev"] | None = None,
    ) -> "ModelConfig":
        return ConfigResolution.resolve(model_name=model_name, base_model=base_model)


AVAILABLE_MODELS = {
    "dev": ModelConfig(
        priority=0,
        aliases=["dev"],
        model_name="black-forest-labs/FLUX.1-dev",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=True,
        requires_sigma_shift=True,
    ),
    "schnell": ModelConfig(
        priority=1,
        aliases=["schnell"],
        model_name="black-forest-labs/FLUX.1-schnell",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=256,
        supports_guidance=False,
        requires_sigma_shift=False,
    ),
    "dev-kontext": ModelConfig(
        priority=2,
        aliases=["dev-kontext"],
        model_name="black-forest-labs/FLUX.1-Kontext-dev",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=True,
        requires_sigma_shift=True,
    ),
    "dev-fill": ModelConfig(
        priority=3,
        aliases=["dev-fill"],
        model_name="black-forest-labs/FLUX.1-Fill-dev",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=True,
        requires_sigma_shift=True,
    ),
    "dev-redux": ModelConfig(
        priority=4,
        aliases=["dev-redux"],
        model_name="black-forest-labs/FLUX.1-Redux-dev",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=True,
        requires_sigma_shift=True,
    ),
    "dev-depth": ModelConfig(
        priority=5,
        aliases=["dev-depth"],
        model_name="black-forest-labs/FLUX.1-Depth-dev",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=True,
        requires_sigma_shift=True,
    ),
    "dev-controlnet-canny": ModelConfig(
        priority=6,
        aliases=["dev-controlnet-canny"],
        model_name="black-forest-labs/FLUX.1-dev",
        base_model=None,
        controlnet_model="InstantX/FLUX.1-dev-Controlnet-Canny",
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=True,
        requires_sigma_shift=True,
    ),
    "schnell-controlnet-canny": ModelConfig(
        priority=7,
        aliases=["schnell-controlnet-canny"],
        model_name="black-forest-labs/FLUX.1-schnell",
        base_model=None,
        controlnet_model="InstantX/FLUX.1-dev-Controlnet-Canny",
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=256,
        supports_guidance=False,
        requires_sigma_shift=False,
    ),
    "dev-controlnet-upscaler": ModelConfig(
        priority=8,
        aliases=["dev-controlnet-upscaler"],
        model_name="black-forest-labs/FLUX.1-dev",
        base_model=None,
        controlnet_model="jasperai/Flux.1-dev-Controlnet-Upscaler",
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=False,
        requires_sigma_shift=False,
    ),
    "dev-fill-catvton": ModelConfig(
        priority=9,
        aliases=["dev-fill-catvton"],
        model_name="black-forest-labs/FLUX.1-Fill-dev",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model="xiaozaa/catvton-flux-beta",
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=True,
        requires_sigma_shift=False,
    ),
    "krea-dev": ModelConfig(
        priority=10,
        aliases=["krea-dev", "dev-krea"],
        model_name="black-forest-labs/FLUX.1-Krea-dev",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=True,
        requires_sigma_shift=True,
    ),
    "qwen-image": ModelConfig(
        priority=11,
        aliases=["qwen-image", "qwen"],
        model_name="Qwen/Qwen-Image",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=None,
        max_sequence_length=None,
        supports_guidance=None,
        requires_sigma_shift=None,
    ),
    "qwen-image-edit": ModelConfig(
        priority=12,
        aliases=["qwen-image-edit", "qwen-edit", "qwen-edit-plus", "qwen-edit-2509"],
        model_name="Qwen/Qwen-Image-Edit-2509",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=None,
        max_sequence_length=None,
        supports_guidance=None,
        requires_sigma_shift=None,
    ),
    "fibo": ModelConfig(
        priority=13,
        aliases=["fibo"],
        model_name="briaai/FIBO",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=True,
        requires_sigma_shift=False,
    ),
    "z-image-turbo": ModelConfig(
        priority=14,
        aliases=["z-image-turbo", "z-image", "zimage-turbo", "zimage"],
        model_name="Tongyi-MAI/Z-Image-Turbo",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=1000,
        max_sequence_length=512,
        supports_guidance=False,  # Turbo model uses guidance_scale=0
        requires_sigma_shift=True,
    ),
    "seedvr2-3b": ModelConfig(
        priority=15,
        aliases=["seedvr2-3b", "seedvr2"],
        model_name="numz/SeedVR2_comfyUI",
        base_model=None,
        controlnet_model=None,
        custom_transformer_model=None,
        num_train_steps=None,
        max_sequence_length=None,
        supports_guidance=True,
        requires_sigma_shift=None,
    ),
}



import mlx.core as mx
from mlx import nn

from mflux.models.common.config.config import Config
from mflux.models.fibo.model.fibo_transformer.fibo_embed_nd import FiboEmbedND
from mflux.models.fibo.model.fibo_transformer.joint_transformer_block import FiboJointTransformerBlock
from mflux.models.fibo.model.fibo_transformer.single_transformer_block import FiboSingleTransformerBlock
from mflux.models.fibo.model.fibo_transformer.text_projection import BriaFiboTextProjection
from mflux.models.fibo.model.fibo_transformer.time_embed import BriaFiboTimestepProjEmbeddings
from mflux.models.flux.model.flux_transformer.ada_layer_norm_continuous import AdaLayerNormContinuous


class FiboTransformer(nn.Module):
    def __init__(
        self,
        in_channels: int = 48,
        num_layers: int = 8,
        num_single_layers: int = 38,
    ):
        super().__init__()
        self.pos_embed = FiboEmbedND()
        self.x_embedder = nn.Linear(in_channels, 3072)
        self.time_embed = BriaFiboTimestepProjEmbeddings()
        self.context_embedder = nn.Linear(4096, 3072)
        self.transformer_blocks = [FiboJointTransformerBlock(i) for i in range(num_layers)]
        self.single_transformer_blocks = [FiboSingleTransformerBlock(i) for i in range(num_single_layers)]
        self.norm_out = AdaLayerNormContinuous(3072, 3072)
        self.proj_out = nn.Linear(3072, in_channels)
        self.caption_projection = [BriaFiboTextProjection() for _ in range(num_layers + num_single_layers)]

    def __call__(
        self,
        t: int,
        config: Config,
        hidden_states: mx.array,
        encoder_hidden_states: mx.array,
        text_encoder_layers: list[mx.array],
    ) -> mx.array:
        # 1. Create embeddings
        hidden_states = FiboTransformer._handle_classifier_free_guidance(hidden_states, encoder_hidden_states)
        hidden_states = self.x_embedder(hidden_states)
        encoder_hidden_states = self.context_embedder(encoder_hidden_states)
        time_embeddings = FiboTransformer._compute_time_embeddings(t, config, hidden_states.shape[0], hidden_states.dtype, self.time_embed)  # fmt: off
        image_rotary_emb = FiboTransformer._compute_rotary_embeddings(encoder_hidden_states, self.pos_embed, config, hidden_states.dtype)  # fmt: off

        # 2. Compute attention mask
        attention_mask = FiboTransformer._compute_attention_mask(
            config=config,
            batch_size=hidden_states.shape[0],
            encoder_hidden_states=encoder_hidden_states,
            max_tokens=encoder_hidden_states.shape[1],
        )

        # 3. Project the fibo-specific text encoder layers
        text_encoder_layers = [
            self.caption_projection[i](text_layer)
            for i, text_layer in enumerate(text_encoder_layers)
        ]  # fmt: off

        # 4. Run the joint transformer blocks
        block_id = 0
        for _, block in enumerate(self.transformer_blocks):
            encoder_hidden_states, hidden_states = FiboTransformer._apply_joint_transformer_block(
                block=block,
                time_embeddings=time_embeddings,
                hidden_states=hidden_states,
                encoder_hidden_states=encoder_hidden_states,
                text_encoder_layer=text_encoder_layers[block_id],
                image_rotary_emb=image_rotary_emb,
                attention_mask=attention_mask,
            )
            block_id += 1

        # 5. Run the single transformer blocks
        for _, block in enumerate(self.single_transformer_blocks):
            encoder_hidden_states, hidden_states = FiboTransformer._apply_single_transformer_block(
                block=block,
                time_embeddings=time_embeddings,
                hidden_states=hidden_states,
                encoder_hidden_states=encoder_hidden_states,
                text_encoder_layer=text_encoder_layers[block_id],
                image_rotary_emb=image_rotary_emb,
                attention_mask=attention_mask,
            )
            block_id += 1

        # 6. Project the final output
        hidden_states = self.norm_out(hidden_states, time_embeddings)
        hidden_states = self.proj_out(hidden_states)
        return hidden_states

    @staticmethod
    def _apply_joint_transformer_block(
        time_embeddings: mx.array,
        block: FiboJointTransformerBlock,
        hidden_states: mx.array,
        encoder_hidden_states: mx.array,
        text_encoder_layer: mx.array,
        image_rotary_emb: mx.array,
        attention_mask: mx.array,
    ) -> tuple[mx.array, mx.array]:
        encoder_hidden_states_half = encoder_hidden_states[:, :, :1536]
        encoder_hidden_states = mx.concatenate([encoder_hidden_states_half, text_encoder_layer], axis=-1)

        encoder_hidden_states, hidden_states = block(
            temb=time_embeddings,
            hidden_states=hidden_states,
            encoder_hidden_states=encoder_hidden_states,
            image_rotary_emb=image_rotary_emb,
            attention_mask=attention_mask,
        )

        return encoder_hidden_states, hidden_states

    @staticmethod
    def _apply_single_transformer_block(
        time_embeddings: mx.array,
        block: FiboSingleTransformerBlock,
        hidden_states: mx.array,
        encoder_hidden_states: mx.array,
        text_encoder_layer: mx.array,
        image_rotary_emb: mx.array,
        attention_mask: mx.array,
    ) -> tuple[mx.array, mx.array]:
        encoder_hidden_states_half = encoder_hidden_states[:, :, :1536]
        encoder_hidden_states = mx.concatenate([encoder_hidden_states_half, text_encoder_layer], axis=-1)
        combined = mx.concatenate([encoder_hidden_states, hidden_states], axis=1)

        combined = block(
            temb=time_embeddings,
            hidden_states=combined,
            image_rotary_emb=image_rotary_emb,
            attention_mask=attention_mask,
        )

        encoder_len = encoder_hidden_states.shape[1]
        encoder_hidden_states = combined[:, :encoder_len, ...]
        hidden_states = combined[:, encoder_len:, ...]

        return encoder_hidden_states, hidden_states

    @staticmethod
    def _handle_classifier_free_guidance(hidden_states: mx.array, encoder_hidden_states: mx.array) -> mx.array:
        batch_size = hidden_states.shape[0]
        encoder_batch_size = encoder_hidden_states.shape[0]
        if encoder_batch_size == batch_size * 2:
            hidden_states = mx.concatenate([hidden_states, hidden_states], axis=0)
        return hidden_states

    @staticmethod
    def _compute_time_embeddings(
        t: int,
        config: Config,
        batch_size: int,
        dtype: mx.Dtype,
        time_embed: BriaFiboTimestepProjEmbeddings,
    ) -> mx.array:
        timestep_value = config.scheduler.timesteps[t]
        timestep = mx.full((batch_size,), timestep_value, dtype=dtype)
        return time_embed(timestep, dtype=dtype)

    @staticmethod
    def _compute_rotary_embeddings(
        encoder_hidden_states: mx.array,
        pos_embed: FiboEmbedND,
        config: Config,
        dtype: mx.Dtype,
    ) -> mx.array:
        max_tokens = encoder_hidden_states.shape[1]
        txt_ids = mx.zeros((max_tokens, 3), dtype=dtype)
        img_ids = FiboTransformer._prepare_latent_image_ids(height=config.height, width=config.width, dtype=dtype)

        if txt_ids.ndim == 3 and txt_ids.shape[0] == 1:
            txt_ids = txt_ids[0]
        if img_ids.ndim == 3 and img_ids.shape[0] == 1:
            img_ids = img_ids[0]

        ids = mx.concatenate((txt_ids, img_ids), axis=0)
        ids = mx.expand_dims(ids, axis=0)
        return pos_embed(ids)

    @staticmethod
    def _prepare_latent_image_ids(height: int, width: int, dtype=mx.float32) -> mx.array:
        vae_scale_factor = 16
        latent_height = height // vae_scale_factor
        latent_width = width // vae_scale_factor
        row_indices = mx.arange(0, latent_height, dtype=dtype)[:, None]
        row_indices = mx.broadcast_to(row_indices, (latent_height, latent_width))
        col_indices = mx.arange(0, latent_width, dtype=dtype)[None, :]
        col_indices = mx.broadcast_to(col_indices, (latent_height, latent_width))
        zeros_channel = mx.zeros((latent_height, latent_width), dtype=dtype)
        latent_image_ids = mx.stack([zeros_channel, row_indices, col_indices], axis=-1)
        latent_image_ids = mx.reshape(latent_image_ids, (latent_height * latent_width, 3))
        return latent_image_ids

    @staticmethod
    def _prepare_attention_mask(attention_mask_2d: mx.array) -> mx.array:
        attention_matrix = mx.einsum("bi,bj->bij", attention_mask_2d, attention_mask_2d)
        mask_dtype = attention_mask_2d.dtype
        min_dtype_value = mx.finfo(mask_dtype).min
        attention_matrix = mx.where(
            attention_matrix == 1,
            mx.zeros_like(attention_matrix).astype(mask_dtype),
            (mx.ones_like(attention_matrix) * min_dtype_value).astype(mask_dtype),
        )
        attention_matrix = mx.expand_dims(attention_matrix, axis=1)
        return attention_matrix

    @staticmethod
    def _compute_attention_mask(
        batch_size: int,
        config: Config,
        encoder_hidden_states: mx.array,
        max_tokens: int,
    ) -> mx.array:
        vae_scale_factor = 16
        latent_height = config.height // vae_scale_factor
        latent_width = config.width // vae_scale_factor
        latent_seq_len = latent_height * latent_width
        prompt_attention_mask = mx.ones((batch_size, max_tokens), dtype=mx.float32)
        latent_attention_mask = mx.ones((batch_size, latent_seq_len), dtype=mx.float32)
        attention_mask_2d = mx.concatenate([prompt_attention_mask, latent_attention_mask], axis=1)
        attention_mask = FiboTransformer._prepare_attention_mask(attention_mask_2d)
        attention_mask = attention_mask.astype(encoder_hidden_states.dtype)
        return attention_mask



import mlx.core as mx
from mlx import nn


class PatchMerger(nn.Module):
    def __init__(self, context_dim: int, hidden_size: int, spatial_merge_size: int = 2):
        super().__init__()
        self.spatial_merge_size = spatial_merge_size
        self.hidden_size_merged = context_dim * (spatial_merge_size**2)
        self.ln_q = nn.RMSNorm(context_dim, eps=1e-6)
        self.mlp_0 = nn.Linear(self.hidden_size_merged, self.hidden_size_merged, bias=True)
        self.mlp_1 = nn.Linear(self.hidden_size_merged, hidden_size, bias=True)

    def __call__(self, x: mx.array, grid_thw: mx.array) -> mx.array:
        if not hasattr(self, "_weights_logged"):
            self._weights_logged = True
        x = self.ln_q(x)
        merged_patches = []
        offset = 0
        for t, h, w in grid_thw:
            t, h, w = int(t), int(h), int(w)
            num_patches_this_image = t * h * w
            x_this_image = x[offset : offset + num_patches_this_image]
            x_merged = x_this_image.reshape(-1, self.hidden_size_merged)
            merged_patches.append(x_merged)
            offset += num_patches_this_image

        x = mx.concatenate(merged_patches, axis=0)
        x = self.mlp_0(x)
        x = nn.gelu(x)
        x = self.mlp_1(x)
        return x



import mlx.core as mx
from mlx import nn

from mflux.models.seedvr2.model.seedvr2_vae.common.conv3d import CausalConv3d


class Downsample3D(nn.Module):
    def __init__(
        self,
        channels: int,
        spatial_only: bool = False,
    ):
        super().__init__()
        kt, st, pt = (1, 1, 0) if spatial_only else (3, 2, 1)
        self.conv = CausalConv3d(
            channels,
            channels,
            kernel_size=(kt, 3, 3),
            stride=(st, 2, 2),
            padding=(pt, 0, 0),
        )

    def __call__(self, x: mx.array) -> mx.array:
        x = mx.pad(x, [(0, 0), (0, 0), (0, 0), (0, 1), (0, 1)])
        return self.conv(x)



import mlx.core as mx
import numpy as np
from mlx import nn


class Qwen3VLVisionRotaryEmbedding(nn.Module):
    def __init__(self, dim: int, theta: float = 10000.0):
        super().__init__()
        inv_freq = 1.0 / (theta ** (np.arange(0, dim, 2, dtype=np.float32) / dim))
        self.inv_freq = mx.array(inv_freq)

    def __call__(self, seqlen: int) -> mx.array:
        seq = mx.arange(seqlen, dtype=mx.float32)
        freqs = mx.outer(seq, self.inv_freq)
        return freqs



import mlx.core as mx
from mlx import nn

from mflux.models.fibo.model.fibo_transformer.bria_fibo_timesteps import BriaFiboTimesteps
from mflux.models.flux.model.flux_transformer.timestep_embedder import TimestepEmbedder


class BriaFiboTimestepProjEmbeddings(nn.Module):
    def __init__(self, embedding_dim: int = 3072, time_theta: int = 10000):
        super().__init__()
        self.time_proj = BriaFiboTimesteps(num_channels=256, flip_sin_to_cos=True, downscale_freq_shift=0, time_theta=time_theta, scale=1.0)  # fmt: off
        self.timestep_embedder = TimestepEmbedder()

    def __call__(self, timestep: mx.array, dtype) -> mx.array:
        timesteps_proj = self.time_proj(timestep)
        timesteps_emb = self.timestep_embedder(timesteps_proj.astype(dtype))
        return timesteps_emb



import mlx.core as mx
from mlx import nn

from mflux.models.fibo.model.fibo_transformer.feed_forward import FiboFeedForward
from mflux.models.fibo.model.fibo_transformer.fibo_ada_layer_norm_zero import FiboAdaLayerNormZero
from mflux.models.fibo.model.fibo_transformer.fibo_joint_attention import FiboJointAttention


class FiboJointTransformerBlock(nn.Module):
    def __init__(
        self,
        layer: int,
        dim: int = 3072,
        num_attention_heads: int = 24,
        attention_head_dim: int = 128,
        eps: float = 1e-6,
    ):
        super().__init__()
        self.layer = layer
        self.norm1 = FiboAdaLayerNormZero(embedding_dim=dim)
        self.norm1_context = FiboAdaLayerNormZero(embedding_dim=dim)
        self.attn = FiboJointAttention(dim=dim, num_attention_heads=num_attention_heads, attention_head_dim=attention_head_dim)  # fmt: off
        self.norm2 = nn.LayerNorm(dims=dim, eps=eps, affine=False)
        self.ff = FiboFeedForward(dim=dim, dim_out=dim, mult=4, activation_fn="gelu-approximate")
        self.norm2_context = nn.LayerNorm(dims=dim, eps=eps, affine=False)
        self.ff_context = FiboFeedForward(dim=dim, dim_out=dim, mult=4, activation_fn="gelu-approximate")

    def __call__(
        self,
        temb: mx.array,
        hidden_states: mx.array,
        encoder_hidden_states: mx.array,
        image_rotary_emb: mx.array,
        attention_mask: mx.array | None = None,
    ) -> tuple[mx.array, mx.array]:
        # 1. AdaLayerNormZero for both streams
        norm_hidden_states, gate_msa, shift_mlp, scale_mlp, gate_mlp = self.norm1(
            hidden_states=hidden_states,
            text_embeddings=temb,
        )

        norm_encoder_hidden_states, c_gate_msa, c_shift_mlp, c_scale_mlp, c_gate_mlp = self.norm1_context(
            hidden_states=encoder_hidden_states,
            text_embeddings=temb,
        )

        # 2. Joint attention over image + context streams.
        attn_output, context_attn_output = self.attn(
            hidden_states=norm_hidden_states,
            encoder_hidden_states=norm_encoder_hidden_states,
            image_rotary_emb=image_rotary_emb,
            attention_mask=attention_mask,
        )

        # 3a. Process attention outputs for the image stream.
        attn_output = mx.expand_dims(gate_msa, axis=1) * attn_output
        hidden_states = hidden_states + attn_output

        norm_hidden_states = self.norm2(hidden_states)
        norm_hidden_states = norm_hidden_states * (1 + scale_mlp[:, None]) + shift_mlp[:, None]

        ff_output = self.ff(norm_hidden_states)
        ff_output = mx.expand_dims(gate_mlp, axis=1) * ff_output

        hidden_states = hidden_states + ff_output

        # 3b. Process attention outputs for the context stream.
        context_attn_output = mx.expand_dims(c_gate_msa, axis=1) * context_attn_output
        encoder_hidden_states = encoder_hidden_states + context_attn_output

        norm_encoder_hidden_states = self.norm2_context(encoder_hidden_states)
        norm_encoder_hidden_states = norm_encoder_hidden_states * (1 + c_scale_mlp[:, None]) + c_shift_mlp[:, None]

        context_ff_output = self.ff_context(norm_encoder_hidden_states)
        encoder_hidden_states = encoder_hidden_states + mx.expand_dims(c_gate_mlp, axis=1) * context_ff_output
        return encoder_hidden_states, hidden_states



import mlx.core as mx


class FiboLatentCreator:
    @staticmethod
    def create_noise(seed: int, height: int, width: int) -> mx.array:
        latents = mx.random.normal(
            shape=(1, 48, (height // 16), (width // 16)),
            key=mx.random.key(seed),
        )
        return FiboLatentCreator.pack_latents(latents, height, width)

    @staticmethod
    def pack_latents(latents: mx.array, height: int, width: int) -> mx.array:
        if latents.ndim == 5:
            latents = latents[:, :, 0, :, :]
        batch_size, channels, latent_height, latent_width = latents.shape
        latents = mx.transpose(latents, (0, 2, 3, 1))
        return mx.reshape(latents, (batch_size, latent_height * latent_width, channels))

    @staticmethod
    def unpack_latents(latents: mx.array, height: int, width: int) -> mx.array:
        batch_size, seq_len, channels = latents.shape
        vae_scale_factor = 16
        latent_height = height // vae_scale_factor
        latent_width = width // vae_scale_factor
        latents = mx.reshape(latents, (batch_size, latent_height, latent_width, channels))
        latents = mx.transpose(latents, (0, 3, 1, 2))
        return latents



import mlx.core as mx
from mlx import nn

from mflux.models.flux.model.flux_vae.common.unet_mid_block import UnetMidBlock
from mflux.models.flux.model.flux_vae.decoder.conv_in import ConvIn
from mflux.models.flux.model.flux_vae.decoder.conv_norm_out import ConvNormOut
from mflux.models.flux.model.flux_vae.decoder.conv_out import ConvOut
from mflux.models.flux.model.flux_vae.decoder.up_block_1_or_2 import UpBlock1Or2
from mflux.models.flux.model.flux_vae.decoder.up_block_3 import UpBlock3
from mflux.models.flux.model.flux_vae.decoder.up_block_4 import UpBlock4


class Decoder(nn.Module):
    def __init__(self, enable_tiling: bool = False, split_direction: str = "horizontal"):
        super().__init__()
        self.conv_in = ConvIn()
        self.mid_block = UnetMidBlock()
        self.up_blocks = [
            UpBlock1Or2(),
            UpBlock1Or2(),
            UpBlock3(),
            UpBlock4(),
        ]
        self.conv_norm_out = ConvNormOut()
        self.conv_out = ConvOut()
        self.enable_tiling = enable_tiling
        self.split_direction = split_direction

    def __call__(self, latents: mx.array) -> mx.array:
        latents = self.conv_in(latents)
        latents = self.mid_block(latents)
        for i, up_block in enumerate(self.up_blocks):
            latents = self._apply_up_block(i, up_block, latents)
        latents = self.conv_norm_out(latents)
        latents = nn.silu(latents)
        latents = self.conv_out(latents)
        return latents

    def _apply_up_block(
        self,
        block_id: int,
        up_block: nn.Module,
        latents: mx.array,
    ) -> mx.array:
        if self.enable_tiling:
            return Decoder._apply_up_block_with_tiling(block_id, up_block, latents, self.split_direction)
        else:
            return up_block(latents)

    @staticmethod
    def _apply_up_block_with_tiling(
        block_id: int,
        up_block: nn.Module,
        latents: mx.array,
        split_direction: str = "horizontal",
    ) -> mx.array:
        # Turns out that the third block is the first to cause OOM issues for larger resolutions.
        # This trick is a "lossy optimization" that splits the latents and processes them separately.
        # The resulting image may have a slightly visible seam at the split point.
        if block_id == 2:
            latents = Decoder._process_block_3_in_tiles(latents, up_block, split_direction)
        else:
            latents = up_block(latents)
        return latents

    @staticmethod
    def _process_block_3_in_tiles(
        latents: mx.array, up_block: nn.Module, split_direction: str = "horizontal"
    ) -> mx.array:
        B, C, H, W = latents.shape
        if split_direction == "horizontal":
            return Decoder._process_horizontal(H, latents, up_block)
        else:
            return Decoder._process_vertical(W, latents, up_block)

    @staticmethod
    def _process_vertical(width: int, latents: mx.array, up_block: nn.Module) -> mx.array:
        # 1. Tile left and right
        left_tile_input = latents[:, :, :, : width // 2]
        right_tile_input = latents[:, :, :, width // 2 :]

        # 2. Process each tile individually
        processed_left_tile = up_block(left_tile_input)
        processed_right_tile = up_block(right_tile_input)

        # 3. Concatenate along width (axis=3)
        return mx.concatenate([processed_left_tile, processed_right_tile], axis=3)

    @staticmethod
    def _process_horizontal(height: int, latents: mx.array, up_block: nn.Module) -> mx.array:
        # 1. Tile top and bottom
        top_tile_input = latents[:, :, : height // 2, :]
        bottom_tile_input = latents[:, :, height // 2 :, :]

        # 2. Process each tile individually
        processed_top_tile = up_block(top_tile_input)
        processed_bottom_tile = up_block(bottom_tile_input)

        # 3. Concatenate along height (axis=2)
        return mx.concatenate([processed_top_tile, processed_bottom_tile], axis=2)



import json
from pathlib import Path
from typing import TYPE_CHECKING, Any

import mlx.core as mx
from mlx import nn
from mlx.utils import tree_flatten
from tqdm import tqdm
from transformers import PreTrainedTokenizer

from mflux.models.common.lora.mapping.lora_saver import LoRASaver
from mflux.utils.version_util import VersionUtil

if TYPE_CHECKING:
    from mflux.models.common.weights.loading.weight_definition import WeightDefinitionType


class ModelSaver:
    @staticmethod
    def save_model(
        model: Any,
        bits: int,
        base_path: str,
        weight_definition: "WeightDefinitionType",
    ) -> None:
        # Save tokenizers from model.tokenizers dict
        tokenizer_defs = weight_definition.get_tokenizers()
        for t in tokenizer_defs:
            if hasattr(model, "tokenizers") and t.name in model.tokenizers:
                tokenizer_wrapper = model.tokenizers[t.name]
                if hasattr(tokenizer_wrapper, "tokenizer"):
                    ModelSaver._save_tokenizer(base_path, tokenizer_wrapper.tokenizer, t.hf_subdir)

        # Save model components with progress bar
        components = [(c.model_attr or c.name, c.hf_subdir) for c in weight_definition.get_components()]
        for attr_name, subdir in tqdm(components, desc="Saving components", unit="component"):
            component = getattr(model, attr_name, None)
            if component is not None:
                # Bake and strip any LoRA wrappers to avoid duplicating shared weights
                LoRASaver.bake_and_strip_lora(component)
                ModelSaver._save_weights(base_path, bits, component, subdir)

    @staticmethod
    def _save_tokenizer(base_path: str, tokenizer: PreTrainedTokenizer, subdir: str) -> None:
        path = Path(base_path) / subdir
        path.mkdir(parents=True, exist_ok=True)
        tokenizer.save_pretrained(path)

    @staticmethod
    def _save_weights(base_path: str, bits: int, model: nn.Module, subdir: str) -> None:
        path = Path(base_path) / subdir
        path.mkdir(parents=True, exist_ok=True)
        weights = dict(tree_flatten(model.parameters()))
        shards = ModelSaver._split_weights(weights)

        # Build weight_map for index.json (maps each weight key to its shard file)
        weight_map = {}
        shard_iter = tqdm(enumerate(shards), total=len(shards), desc=f"  {subdir}", unit="shard", leave=False)
        for i, shard in shard_iter:
            shard_filename = f"{i}.safetensors"
            mx.save_safetensors(
                str(path / shard_filename),
                shard,
                {
                    "quantization_level": str(bits),
                    "mflux_version": VersionUtil.get_mflux_version(),
                },
            )
            # Record which file each weight belongs to
            for key in shard.keys():
                weight_map[key] = shard_filename

        # Write model.safetensors.index.json for HuggingFace compatibility
        # This ensures the saved model works even if custom metadata is stripped
        index_data = {
            "metadata": {
                "quantization_level": str(bits),
                "mflux_version": VersionUtil.get_mflux_version(),
            },
            "weight_map": weight_map,
        }
        with open(path / "model.safetensors.index.json", "w") as f:
            json.dump(index_data, f, indent=2)

    @staticmethod
    def _split_weights(weights: dict, max_file_size_gb: int = 2) -> list[dict]:
        max_file_size_bytes = max_file_size_gb << 30
        shards: list[dict] = []
        shard: dict = {}
        shard_size = 0
        for k, v in weights.items():
            if shard_size + v.nbytes > max_file_size_bytes:
                shards.append(shard)
                shard, shard_size = {}, 0
            shard[k] = v
            shard_size += v.nbytes
        if shard:  # Don't append empty shard
            shards.append(shard)
        return shards



import os
from pathlib import Path

import cv2
import numpy as np
from PIL import Image


class ImageCompare:
    DEFAULT_MISMATCH_THRESHOLD = 0.15  # Set by eye test
    ENV_MISMATCH_THRESHOLD = float(os.environ.get("MFLUX_IMAGE_MISMATCH_THRESHOLD", DEFAULT_MISMATCH_THRESHOLD))

    @staticmethod
    def check_images_close_enough(
        image1_path: str | Path,
        image2_path: str | Path,
        error_message_prefix: str,
        mismatch_threshold: float | None = None,
    ) -> float:
        if mismatch_threshold is None:
            mismatch_threshold = ImageCompare.ENV_MISMATCH_THRESHOLD

        image1_path = Path(image1_path)
        image2_path = Path(image2_path)
        image1_data = np.array(Image.open(image1_path))
        image2_data = np.array(Image.open(image2_path))
        closeness_array = np.isclose(
            image1_data,
            image2_data,
            rtol=float(os.environ.get("MFLUX_IMAGE_ALLCLOSE_RTOL", 0.1)),
            atol=0,  # ignore absolute tolerance
        )
        num_mismatched = np.count_nonzero(~closeness_array)
        total_elements = image1_data.size
        mismatch_ratio = num_mismatched / total_elements
        if mismatch_ratio > mismatch_threshold:
            diff = image1_data - image2_data
            # Scale the difference so it's visible. A difference of 50 will become bright.
            # We'll scale it so that a difference of 50 or more becomes pure white (255).
            diff_visual = (np.abs(diff) / 50 * 255).clip(0, 255).astype(np.uint8)
            cv2.imwrite(image2_path.with_stem(image1_path.stem + "_diff"), diff_visual)
            raise AssertionError(
                f"{error_message_prefix} Check {image1_path} vs {image2_path} :: their elements are {mismatch_ratio:.1%} different. Fails assertion for {mismatch_threshold=}"
            )
        return mismatch_ratio

    @staticmethod
    def compare_images(
        image1_path: str | Path,
        image2_path: str | Path,
        mismatch_threshold: float | None = None,
        create_diff: bool = True,
    ) -> dict:
        image1_path = Path(image1_path)
        image2_path = Path(image2_path)
        image1_data = np.array(Image.open(image1_path))
        image2_data = np.array(Image.open(image2_path))

        rtol = float(os.environ.get("MFLUX_IMAGE_ALLCLOSE_RTOL", 0.1))
        closeness_array = np.isclose(
            image1_data,
            image2_data,
            rtol=rtol,
            atol=0,
        )

        num_mismatched = np.count_nonzero(~closeness_array)
        total_elements = image1_data.size
        mismatch_ratio = num_mismatched / total_elements

        diff_path = None
        if create_diff:
            diff = image1_data - image2_data
            diff_visual = (np.abs(diff) / 50 * 255).clip(0, 255).astype(np.uint8)
            diff_path = image2_path.with_stem(image1_path.stem + "_diff")
            cv2.imwrite(str(diff_path), diff_visual)

        result = {
            "mismatch_ratio": mismatch_ratio,
            "total_pixels": total_elements,
            "mismatched_pixels": num_mismatched,
            "image1_shape": image1_data.shape,
            "image2_shape": image2_data.shape,
            "diff_path": str(diff_path) if diff_path else None,
        }

        if mismatch_threshold is not None:
            result["passes_threshold"] = mismatch_ratio <= mismatch_threshold
            result["threshold"] = mismatch_threshold

        return result



import mlx.core as mx
from mlx import nn


class QwenImageCausalConv3D(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 1,
        padding: int = 1,
    ):
        super().__init__()
        self.conv3d = nn.Conv3d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=0,
        )
        self.padding = padding
        self.stride = stride
        self.kernel_size = kernel_size

    def __call__(self, x: mx.array) -> mx.array:
        pad_t = pad_h = pad_w = self.padding
        if pad_t > 0 or pad_h > 0 or pad_w > 0:
            pad_spec = [
                (0, 0),
                (0, 0),
                (2 * pad_t, 0),
                (pad_h, pad_h),
                (pad_w, pad_w),
            ]
            x = mx.pad(x, pad_spec)

        x = mx.transpose(x, (0, 2, 3, 4, 1))
        x = self.conv3d(x)
        x = mx.transpose(x, (0, 4, 1, 2, 3))
        return x



from pathlib import Path

import mlx.core as mx
from mlx import nn

from mflux.models.common.config.config import Config
from mflux.models.common.config.model_config import ModelConfig
from mflux.models.common.latent_creator.latent_creator import LatentCreator
from mflux.models.common.vae.vae_util import VAEUtil
from mflux.models.flux.flux_initializer import FluxInitializer
from mflux.models.flux.latent_creator.flux_latent_creator import FluxLatentCreator
from mflux.models.flux.model.flux_text_encoder.clip_encoder.clip_encoder import CLIPEncoder
from mflux.models.flux.model.flux_text_encoder.prompt_encoder import PromptEncoder
from mflux.models.flux.model.flux_text_encoder.t5_encoder.t5_encoder import T5Encoder
from mflux.models.flux.model.flux_transformer.transformer import Transformer
from mflux.models.flux.model.flux_vae.vae import VAE
from mflux.utils.exceptions import StopImageGenerationException
from mflux.utils.generated_image import GeneratedImage
from mflux.utils.image_util import ImageUtil


class Flux1InContextDev(nn.Module):
    vae: VAE
    transformer: Transformer
    t5_text_encoder: T5Encoder
    clip_text_encoder: CLIPEncoder

    def __init__(
        self,
        quantize: int | None = None,
        model_path: str | None = None,
        lora_paths: list[str] | None = None,
        lora_scales: list[float] | None = None,
        model_config: ModelConfig = ModelConfig.dev(),
    ):
        super().__init__()
        FluxInitializer.init(
            model=self,
            quantize=quantize,
            model_path=model_path,
            lora_paths=lora_paths,
            lora_scales=lora_scales,
            model_config=model_config,
        )

    def generate_image(
        self,
        seed: int,
        prompt: str,
        num_inference_steps: int = 4,
        height: int = 1024,
        width: int = 1024,
        guidance: float = 4.0,
        image_path: Path | str | None = None,
        image_strength: float | None = None,
        scheduler: str = "linear",
    ) -> GeneratedImage:
        # 0. Create a new config based on the model type and input parameters
        config = Config(
            width=width,
            height=height,
            guidance=guidance,
            scheduler=scheduler,
            image_path=image_path,
            image_strength=image_strength,
            model_config=self.model_config,
            num_inference_steps=num_inference_steps,
        )

        # 1. Encode the reference image
        encoded_image = LatentCreator.encode_image(
            vae=self.vae,
            width=config.width,
            height=config.height,
            image_path=config.image_path,
            tiling_config=self.tiling_config,
        )

        # 2. Create the initial latents and keep the initial static noise for later blending
        static_noise = Flux1InContextDev._create_in_context_latents(seed=seed, config=config)
        latents = mx.array(static_noise)

        # 3. Encode the prompt
        prompt_embeds, pooled_prompt_embeds = PromptEncoder.encode_prompt(
            prompt=prompt,
            prompt_cache=self.prompt_cache,
            t5_tokenizer=self.tokenizers["t5"],
            clip_tokenizer=self.tokenizers["clip"],
            t5_text_encoder=self.t5_text_encoder,
            clip_text_encoder=self.clip_text_encoder,
        )

        # 4. Create callback context and call before_loop
        ctx = self.callbacks.start(seed=seed, prompt=prompt, config=config)
        ctx.before_loop(latents)

        for t in config.time_steps:
            try:
                # Scale model input if needed by the scheduler
                latents = config.scheduler.scale_model_input(latents, t)

                # 5.t Predict the noise
                noise = self.transformer(
                    t=t,
                    config=config,
                    hidden_states=latents,
                    prompt_embeds=prompt_embeds,
                    pooled_prompt_embeds=pooled_prompt_embeds,
                )

                # 6.t Take one denoise step and update latents
                latents = config.scheduler.step(noise=noise, timestep=t, latents=latents)

                # 7.t Override the left-hand side of latents by linearly interpolating between latents and static noise
                latents = Flux1InContextDev._update_latents(
                    t=t,
                    config=config,
                    latents=latents,
                    encoded_image=encoded_image,
                    static_noise=static_noise,
                    sigmas=config.scheduler.sigmas,
                )

                # 8.t Call subscribers in-loop
                ctx.in_loop(t, latents)

                # (Optional) Evaluate to enable progress tracking
                mx.eval(latents)

            except KeyboardInterrupt:  # noqa: PERF203
                ctx.interruption(t, latents)
                raise StopImageGenerationException(
                    f"Stopping image generation at step {t + 1}/{config.num_inference_steps}"
                )

        # 9. Call subscribers after loop
        ctx.after_loop(latents)

        # 10. Decode the latent array and return the image
        latents = FluxLatentCreator.unpack_latents(latents=latents, height=config.height, width=config.width)
        decoded = VAEUtil.decode(vae=self.vae, latent=latents, tiling_config=self.tiling_config)
        return ImageUtil.to_image(
            decoded_latents=decoded,
            config=config,
            seed=seed,
            prompt=prompt,
            quantization=self.bits,
            lora_paths=self.lora_paths,
            lora_scales=self.lora_scales,
            image_path=config.image_path,
            image_strength=config.image_strength,
            generation_time=config.time_steps.format_dict["elapsed"],
        )

    @staticmethod
    def _create_in_context_latents(seed: int, config: Config):
        # 1. Double the width for side-by-side generation
        config.width = 2 * config.width

        # 2. Create the initial latents with doubled width
        latent_height = config.height // 8
        latent_width = config.width // 8

        # 3. Create noise with appropriate dimensions
        static_noise = mx.random.normal(shape=[1, 16, latent_height, latent_width], key=mx.random.key(seed))
        latents = FluxLatentCreator.pack_latents(latents=static_noise, height=config.height, width=config.width)
        return latents

    @staticmethod
    def _update_latents(
        t: int,
        config: Config,
        latents: mx.array,
        encoded_image: mx.array,
        static_noise: mx.array,
        sigmas: mx.array,
    ) -> mx.array:
        # 1. Unpack the latents
        unpacked = FluxLatentCreator.unpack_latents(latents=latents, height=config.height, width=config.width)
        unpacked_static_noise = FluxLatentCreator.unpack_latents(
            latents=static_noise, height=config.height, width=config.width
        )

        # 2. Calculate latent_width from the config (original width is half of current width)
        latent_width = (config.width // 2) // 8

        # 3. Override the left side with the reference image blended with appropriate noise for current timestep
        unpacked[:, :, :, 0:latent_width] = LatentCreator.add_noise_by_interpolation(
            clean=encoded_image[:, :, :, 0:latent_width],
            noise=unpacked_static_noise[:, :, :, 0:latent_width],
            sigma=sigmas[t + 1],
        )

        # 4. Repack the latents
        return FluxLatentCreator.pack_latents(latents=unpacked, height=config.height, width=config.width)



import logging

from mflux.models.common.resolution.actions import QuantizationAction, Rule

logger = logging.getLogger(__name__)


class QuantizationResolution:
    RULES = frozenset(
        {
            Rule(priority=0, name="none", check="none_none", action=QuantizationAction.NONE),
            Rule(priority=1, name="on_the_fly", check="none_any", action=QuantizationAction.REQUESTED),
            Rule(priority=2, name="pre_quantized", check="any_none", action=QuantizationAction.STORED),
            Rule(priority=3, name="conflict", check="any_any", action=QuantizationAction.STORED),
        }
    )

    @staticmethod
    def resolve(stored: int | None, requested: int | None) -> tuple[int | None, str | None]:
        for rule in sorted(QuantizationResolution.RULES, key=lambda r: r.priority):
            if QuantizationResolution._check(rule.check, stored, requested):
                logger.debug(
                    f"Quantization resolution: stored={stored}, requested={requested} "
                    f"→ rule '{rule.name}' ({rule.action.value})"
                )
                return QuantizationResolution._execute(rule, stored, requested)

        raise ValueError(f"Unexpected quantization state: stored={stored}, requested={requested}")

    @staticmethod
    def _check(check: str, stored: int | None, requested: int | None) -> bool:
        if check == "none_none":
            return stored is None and requested is None
        if check == "none_any":
            return stored is None and requested is not None
        if check == "any_none":
            return stored is not None and requested is None
        if check == "any_any":
            return stored is not None and requested is not None
        return False

    @staticmethod
    def _execute(rule: Rule, stored: int | None, requested: int | None) -> tuple[int | None, str | None]:
        if rule.action == QuantizationAction.NONE:
            return None, None
        if rule.action == QuantizationAction.REQUESTED:
            return requested, None
        if rule.action == QuantizationAction.STORED:
            warn = rule.name == "conflict" and stored != requested
            warning = f"Model is pre-quantized at {stored}-bit. Ignoring -q {requested} flag." if warn else None
            return stored, warning
        return None, None



from abc import ABC, abstractmethod
from typing import Protocol, runtime_checkable

import mlx.core as mx
import numpy as np
from PIL import Image
from transformers import PreTrainedTokenizer

from mflux.models.common.tokenizer.tokenizer_output import TokenizerOutput


@runtime_checkable
class Tokenizer(Protocol):
    tokenizer: PreTrainedTokenizer

    def tokenize(
        self,
        prompt: str | list[str],
        images: list[Image.Image] | None = None,
        max_length: int | None = None,
        **kwargs,
    ) -> TokenizerOutput: ...


class BaseTokenizer(ABC):
    def __init__(self, tokenizer: PreTrainedTokenizer, max_length: int = 512):
        self.tokenizer = tokenizer
        self.max_length = max_length

    @abstractmethod
    def tokenize(
        self,
        prompt: str | list[str],
        images: list[Image.Image] | None = None,
        max_length: int | None = None,
        **kwargs,
    ) -> TokenizerOutput: ...


class LanguageTokenizer(BaseTokenizer):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        max_length: int = 512,
        padding: str = "max_length",
        return_attention_mask: bool = True,
        template: str | None = None,
        use_chat_template: bool = False,
        chat_template_kwargs: dict | None = None,
        add_special_tokens: bool = True,
    ):
        super().__init__(tokenizer, max_length)
        self.padding = padding
        self.return_attention_mask = return_attention_mask
        self.template = template
        self.use_chat_template = use_chat_template
        self.chat_template_kwargs = chat_template_kwargs or {}
        self.add_special_tokens = add_special_tokens

    def tokenize(
        self,
        prompt: str | list[str],
        images: list[Image.Image] | None = None,
        max_length: int | None = None,
        **kwargs,
    ) -> TokenizerOutput:
        max_length = max_length or self.max_length

        if isinstance(prompt, str):
            prompts = [prompt]
        else:
            prompts = list(prompt)

        prompts = [p if p is not None else "" for p in prompts]
        if all(p == "" for p in prompts):
            batch_size = len(prompts)
            input_ids = mx.array(np.empty((batch_size, 0), dtype=np.int32))
            attention_mask = mx.array(np.empty((batch_size, 0), dtype=np.int32))
            return TokenizerOutput(input_ids=input_ids, attention_mask=attention_mask)

        if self.template or self.use_chat_template:
            formatted_prompts = []
            for p in prompts:
                if self.template:
                    formatted = self.template.format(p)
                elif self.use_chat_template:
                    formatted = self.tokenizer.apply_chat_template(
                        [{"role": "user", "content": p}],
                        tokenize=False,
                        add_generation_prompt=True,
                        **self.chat_template_kwargs,
                    )
                else:
                    formatted = p
                formatted_prompts.append(formatted)
            prompts = formatted_prompts

        tokens = self.tokenizer(
            prompts,
            padding=self.padding,
            max_length=max_length,
            truncation=True,
            add_special_tokens=self.add_special_tokens,
            return_length=False,
            return_overflowing_tokens=False,
            return_tensors="np",
        )

        input_ids = mx.array(tokens["input_ids"])
        if self.return_attention_mask:
            attention_mask = mx.array(tokens["attention_mask"])
        else:
            attention_mask = mx.ones_like(input_ids)

        return TokenizerOutput(
            input_ids=input_ids,
            attention_mask=attention_mask,
        )


class VisionLanguageTokenizer(BaseTokenizer):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        processor,
        max_length: int = 1024,
        template: str | None = None,
        image_token: str = "<|image_pad|>",
    ):
        super().__init__(tokenizer, max_length)
        self.processor = processor
        self.template = template
        self.image_token = image_token

    def tokenize(
        self,
        prompt: str | list[str],
        images: list[Image.Image] | None = None,
        max_length: int | None = None,
        **kwargs,
    ) -> TokenizerOutput:
        max_length = max_length or self.max_length

        if isinstance(prompt, str):
            prompt = [prompt]

        if self.template and images:
            img_prompt = ""
            for i in range(len(images)):
                img_prompt += f"Picture {i + 1}: <|vision_start|>{self.image_token}<|vision_end|>"
            formatted_text = self.template.format(img_prompt + prompt[0])
        elif self.template:
            formatted_text = self.template.format(prompt[0])
        else:
            formatted_text = prompt[0]

        pixel_values = None
        image_grid_thw = None

        if images:
            model_inputs = self.processor(
                text=[formatted_text],
                images=images,
                padding=True,
                return_tensors=None,
            )

            input_ids = model_inputs["input_ids"]
            attention_mask = model_inputs["attention_mask"]
            pixel_values = mx.array(model_inputs["pixel_values"])
            image_grid_thw = mx.array(model_inputs["image_grid_thw"])
        else:
            tokens = self.tokenizer(
                [formatted_text],
                max_length=max_length,
                padding=True,
                truncation=True,
                return_tensors="np",
            )
            input_ids = mx.array(tokens["input_ids"])
            attention_mask = mx.array(tokens["attention_mask"])

        return TokenizerOutput(
            input_ids=input_ids,
            attention_mask=attention_mask,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
        )



import mlx.core as mx
from mlx import nn

from mflux.models.z_image.model.z_image_vae.common.unet_mid_block import UNetMidBlock
from mflux.models.z_image.model.z_image_vae.decoder.conv_in import ConvIn
from mflux.models.z_image.model.z_image_vae.decoder.conv_norm_out import ConvNormOut
from mflux.models.z_image.model.z_image_vae.decoder.conv_out import ConvOut
from mflux.models.z_image.model.z_image_vae.decoder.up_decoder_block import UpDecoderBlock


class Decoder(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv_in = ConvIn(in_channels=16, out_channels=512)
        self.mid_block = UNetMidBlock(channels=512)
        self.up_blocks = [
            UpDecoderBlock(in_channels=512, out_channels=512, num_layers=3, add_upsample=True),
            UpDecoderBlock(in_channels=512, out_channels=512, num_layers=3, add_upsample=True),
            UpDecoderBlock(in_channels=512, out_channels=256, num_layers=3, add_upsample=True),
            UpDecoderBlock(in_channels=256, out_channels=128, num_layers=3, add_upsample=False),
        ]

        self.conv_norm_out = ConvNormOut(channels=128)
        self.conv_out = ConvOut(in_channels=128, out_channels=3)

    def __call__(self, latents: mx.array) -> mx.array:
        hidden_states = self.conv_in(latents)
        hidden_states = self.mid_block(hidden_states)

        for up_block in self.up_blocks:
            hidden_states = up_block(hidden_states)

        hidden_states = self.conv_norm_out(hidden_states)
        hidden_states = nn.silu(hidden_states)
        hidden_states = self.conv_out(hidden_states)
        return hidden_states



from typing import List

from mflux.models.common.weights.mapping.weight_mapping import WeightMapping, WeightTarget
from mflux.models.common.weights.mapping.weight_transforms import WeightTransforms


class QwenWeightMapping(WeightMapping):
    @staticmethod
    def get_transformer_mapping() -> List[WeightTarget]:
        return [
            WeightTarget(
                to_pattern="img_in.weight",
                from_pattern=["img_in.weight"],
            ),
            WeightTarget(
                to_pattern="img_in.bias",
                from_pattern=["img_in.bias"],
            ),
            WeightTarget(
                to_pattern="txt_norm.weight",
                from_pattern=["txt_norm.weight"],
            ),
            WeightTarget(
                to_pattern="txt_in.weight",
                from_pattern=["txt_in.weight"],
            ),
            WeightTarget(
                to_pattern="txt_in.bias",
                from_pattern=["txt_in.bias"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.timestep_embedder.linear_1.weight",
                from_pattern=["time_text_embed.timestep_embedder.linear_1.weight"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.timestep_embedder.linear_1.bias",
                from_pattern=["time_text_embed.timestep_embedder.linear_1.bias"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.timestep_embedder.linear_2.weight",
                from_pattern=["time_text_embed.timestep_embedder.linear_2.weight"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.timestep_embedder.linear_2.bias",
                from_pattern=["time_text_embed.timestep_embedder.linear_2.bias"],
            ),
            WeightTarget(
                to_pattern="norm_out.linear.weight",
                from_pattern=["norm_out.linear.weight"],
            ),
            WeightTarget(
                to_pattern="norm_out.linear.bias",
                from_pattern=["norm_out.linear.bias"],
            ),
            WeightTarget(
                to_pattern="proj_out.weight",
                from_pattern=["proj_out.weight"],
            ),
            WeightTarget(
                to_pattern="proj_out.bias",
                from_pattern=["proj_out.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_q.weight",
                from_pattern=["transformer_blocks.{block}.attn.to_q.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_q.bias",
                from_pattern=["transformer_blocks.{block}.attn.to_q.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_k.weight",
                from_pattern=["transformer_blocks.{block}.attn.to_k.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_k.bias",
                from_pattern=["transformer_blocks.{block}.attn.to_k.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_v.weight",
                from_pattern=["transformer_blocks.{block}.attn.to_v.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_v.bias",
                from_pattern=["transformer_blocks.{block}.attn.to_v.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_q_proj.weight",
                from_pattern=["transformer_blocks.{block}.attn.add_q_proj.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_q_proj.bias",
                from_pattern=["transformer_blocks.{block}.attn.add_q_proj.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_k_proj.weight",
                from_pattern=["transformer_blocks.{block}.attn.add_k_proj.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_k_proj.bias",
                from_pattern=["transformer_blocks.{block}.attn.add_k_proj.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_v_proj.weight",
                from_pattern=["transformer_blocks.{block}.attn.add_v_proj.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_v_proj.bias",
                from_pattern=["transformer_blocks.{block}.attn.add_v_proj.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.norm_q.weight",
                from_pattern=["transformer_blocks.{block}.attn.norm_q.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.norm_k.weight",
                from_pattern=["transformer_blocks.{block}.attn.norm_k.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.norm_added_q.weight",
                from_pattern=["transformer_blocks.{block}.attn.norm_added_q.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.norm_added_k.weight",
                from_pattern=["transformer_blocks.{block}.attn.norm_added_k.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.attn_to_out.0.weight",
                from_pattern=["transformer_blocks.{block}.attn.to_out.0.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.attn_to_out.0.bias",
                from_pattern=["transformer_blocks.{block}.attn.to_out.0.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_add_out.weight",
                from_pattern=["transformer_blocks.{block}.attn.to_add_out.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_add_out.bias",
                from_pattern=["transformer_blocks.{block}.attn.to_add_out.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.img_mod_linear.weight",
                from_pattern=["transformer_blocks.{block}.img_mod.1.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.img_mod_linear.bias",
                from_pattern=["transformer_blocks.{block}.img_mod.1.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.txt_mod_linear.weight",
                from_pattern=["transformer_blocks.{block}.txt_mod.1.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.txt_mod_linear.bias",
                from_pattern=["transformer_blocks.{block}.txt_mod.1.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.img_ff.mlp_in.weight",
                from_pattern=["transformer_blocks.{block}.img_mlp.net.0.proj.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.img_ff.mlp_in.bias",
                from_pattern=["transformer_blocks.{block}.img_mlp.net.0.proj.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.img_ff.mlp_out.weight",
                from_pattern=["transformer_blocks.{block}.img_mlp.net.2.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.img_ff.mlp_out.bias",
                from_pattern=["transformer_blocks.{block}.img_mlp.net.2.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.txt_ff.mlp_in.weight",
                from_pattern=["transformer_blocks.{block}.txt_mlp.net.0.proj.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.txt_ff.mlp_in.bias",
                from_pattern=["transformer_blocks.{block}.txt_mlp.net.0.proj.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.txt_ff.mlp_out.weight",
                from_pattern=["transformer_blocks.{block}.txt_mlp.net.2.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.txt_ff.mlp_out.bias",
                from_pattern=["transformer_blocks.{block}.txt_mlp.net.2.bias"],
            ),
        ]

    @staticmethod
    def get_vae_mapping() -> List[WeightTarget]:
        return [
            WeightTarget(
                to_pattern="decoder.conv_in.conv3d.weight",
                from_pattern=["decoder.conv_in.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.conv_in.conv3d.bias",
                from_pattern=["decoder.conv_in.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.conv_out.conv3d.weight",
                from_pattern=["decoder.conv_out.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.conv_out.conv3d.bias",
                from_pattern=["decoder.conv_out.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.norm_out.weight",
                from_pattern=["decoder.norm_out.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="post_quant_conv.conv3d.weight",
                from_pattern=["post_quant_conv.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="post_quant_conv.conv3d.bias",
                from_pattern=["post_quant_conv.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv1.conv3d.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv1.conv3d.bias",
                from_pattern=["decoder.mid_block.resnets.{i}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv2.conv3d.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv2.conv3d.bias",
                from_pattern=["decoder.mid_block.resnets.{i}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.norm1.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.norm2.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.norm.weight",
                from_pattern=["decoder.mid_block.attentions.0.norm.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_qkv.weight",
                from_pattern=["decoder.mid_block.attentions.0.to_qkv.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_qkv.bias",
                from_pattern=["decoder.mid_block.attentions.0.to_qkv.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.proj.weight",
                from_pattern=["decoder.mid_block.attentions.0.proj.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.proj.bias",
                from_pattern=["decoder.mid_block.attentions.0.proj.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.resnets.{res}.conv1.conv3d.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.resnets.{res}.conv1.conv3d.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.resnets.{res}.conv2.conv3d.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.resnets.{res}.conv2.conv3d.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.resnets.{res}.norm1.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.resnets.{res}.norm2.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.resnets.{res}.skip_conv.conv3d.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv_shortcut.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.resnets.{res}.skip_conv.conv3d.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv_shortcut.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.upsamplers.0.resample_conv.weight",
                from_pattern=["decoder.up_blocks.{block}.upsamplers.0.resample.1.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.upsamplers.0.resample_conv.bias",
                from_pattern=["decoder.up_blocks.{block}.upsamplers.0.resample.1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.upsamplers.0.time_conv.conv3d.weight",
                from_pattern=["decoder.up_blocks.{block}.upsamplers.0.time_conv.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.up_block{block}.upsamplers.0.time_conv.conv3d.bias",
                from_pattern=["decoder.up_blocks.{block}.upsamplers.0.time_conv.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.conv_in.conv3d.weight",
                from_pattern=["encoder.conv_in.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.conv_in.conv3d.bias",
                from_pattern=["encoder.conv_in.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.conv_out.conv3d.weight",
                from_pattern=["encoder.conv_out.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.conv_out.conv3d.bias",
                from_pattern=["encoder.conv_out.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.norm_out.weight",
                from_pattern=["encoder.norm_out.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.norm.weight",
                from_pattern=["encoder.mid_block.attentions.0.norm.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_qkv.weight",
                from_pattern=["encoder.mid_block.attentions.0.to_qkv.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_qkv.bias",
                from_pattern=["encoder.mid_block.attentions.0.to_qkv.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.proj.weight",
                from_pattern=["encoder.mid_block.attentions.0.proj.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.proj.bias",
                from_pattern=["encoder.mid_block.attentions.0.proj.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv1.conv3d.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv1.conv3d.bias",
                from_pattern=["encoder.mid_block.resnets.{i}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv2.conv3d.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv2.conv3d.bias",
                from_pattern=["encoder.mid_block.resnets.{i}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.norm1.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.norm2.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.0.conv1.conv3d.weight",
                from_pattern=["encoder.down_blocks.0.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.0.conv1.conv3d.bias",
                from_pattern=["encoder.down_blocks.0.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.0.conv2.conv3d.weight",
                from_pattern=["encoder.down_blocks.0.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.0.conv2.conv3d.bias",
                from_pattern=["encoder.down_blocks.0.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.0.norm1.weight",
                from_pattern=["encoder.down_blocks.0.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.0.norm2.weight",
                from_pattern=["encoder.down_blocks.0.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.1.conv1.conv3d.weight",
                from_pattern=["encoder.down_blocks.1.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.1.conv1.conv3d.bias",
                from_pattern=["encoder.down_blocks.1.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.1.conv2.conv3d.weight",
                from_pattern=["encoder.down_blocks.1.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.1.conv2.conv3d.bias",
                from_pattern=["encoder.down_blocks.1.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.1.norm1.weight",
                from_pattern=["encoder.down_blocks.1.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.resnets.1.norm2.weight",
                from_pattern=["encoder.down_blocks.1.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.downsamplers.0.resample_conv.weight",
                from_pattern=["encoder.down_blocks.2.resample.1.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.0.downsamplers.0.resample_conv.bias",
                from_pattern=["encoder.down_blocks.2.resample.1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.0.conv1.conv3d.weight",
                from_pattern=["encoder.down_blocks.3.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.0.conv1.conv3d.bias",
                from_pattern=["encoder.down_blocks.3.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.0.conv2.conv3d.weight",
                from_pattern=["encoder.down_blocks.3.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.0.conv2.conv3d.bias",
                from_pattern=["encoder.down_blocks.3.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.0.norm1.weight",
                from_pattern=["encoder.down_blocks.3.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.0.norm2.weight",
                from_pattern=["encoder.down_blocks.3.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.0.skip_conv.conv3d.weight",
                from_pattern=["encoder.down_blocks.3.conv_shortcut.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.0.skip_conv.conv3d.bias",
                from_pattern=["encoder.down_blocks.3.conv_shortcut.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.1.conv1.conv3d.weight",
                from_pattern=["encoder.down_blocks.4.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.1.conv1.conv3d.bias",
                from_pattern=["encoder.down_blocks.4.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.1.conv2.conv3d.weight",
                from_pattern=["encoder.down_blocks.4.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.1.conv2.conv3d.bias",
                from_pattern=["encoder.down_blocks.4.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.1.norm1.weight",
                from_pattern=["encoder.down_blocks.4.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.resnets.1.norm2.weight",
                from_pattern=["encoder.down_blocks.4.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.downsamplers.0.resample_conv.weight",
                from_pattern=["encoder.down_blocks.5.resample.1.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.1.downsamplers.0.resample_conv.bias",
                from_pattern=["encoder.down_blocks.5.resample.1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.0.conv1.conv3d.weight",
                from_pattern=["encoder.down_blocks.6.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.0.conv1.conv3d.bias",
                from_pattern=["encoder.down_blocks.6.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.0.conv2.conv3d.weight",
                from_pattern=["encoder.down_blocks.6.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.0.conv2.conv3d.bias",
                from_pattern=["encoder.down_blocks.6.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.0.norm1.weight",
                from_pattern=["encoder.down_blocks.6.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.0.norm2.weight",
                from_pattern=["encoder.down_blocks.6.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.0.skip_conv.conv3d.weight",
                from_pattern=["encoder.down_blocks.6.conv_shortcut.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.0.skip_conv.conv3d.bias",
                from_pattern=["encoder.down_blocks.6.conv_shortcut.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.1.conv1.conv3d.weight",
                from_pattern=["encoder.down_blocks.7.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.1.conv1.conv3d.bias",
                from_pattern=["encoder.down_blocks.7.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.1.conv2.conv3d.weight",
                from_pattern=["encoder.down_blocks.7.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.1.conv2.conv3d.bias",
                from_pattern=["encoder.down_blocks.7.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.1.norm1.weight",
                from_pattern=["encoder.down_blocks.7.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.resnets.1.norm2.weight",
                from_pattern=["encoder.down_blocks.7.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.downsamplers.0.resample_conv.weight",
                from_pattern=["encoder.down_blocks.8.resample.1.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.downsamplers.0.resample_conv.bias",
                from_pattern=["encoder.down_blocks.8.resample.1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.downsamplers.0.time_conv.conv3d.weight",
                from_pattern=["encoder.down_blocks.8.time_conv.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.2.downsamplers.0.time_conv.conv3d.bias",
                from_pattern=["encoder.down_blocks.8.time_conv.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.0.conv1.conv3d.weight",
                from_pattern=["encoder.down_blocks.9.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.0.conv1.conv3d.bias",
                from_pattern=["encoder.down_blocks.9.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.0.conv2.conv3d.weight",
                from_pattern=["encoder.down_blocks.9.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.0.conv2.conv3d.bias",
                from_pattern=["encoder.down_blocks.9.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.0.norm1.weight",
                from_pattern=["encoder.down_blocks.9.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.0.norm2.weight",
                from_pattern=["encoder.down_blocks.9.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.1.conv1.conv3d.weight",
                from_pattern=["encoder.down_blocks.10.conv1.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.1.conv1.conv3d.bias",
                from_pattern=["encoder.down_blocks.10.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.1.conv2.conv3d.weight",
                from_pattern=["encoder.down_blocks.10.conv2.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.1.conv2.conv3d.bias",
                from_pattern=["encoder.down_blocks.10.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.1.norm1.weight",
                from_pattern=["encoder.down_blocks.10.norm1.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.3.resnets.1.norm2.weight",
                from_pattern=["encoder.down_blocks.10.norm2.gamma"],
                transform=WeightTransforms.reshape_gamma_to_1d,
            ),
            WeightTarget(
                to_pattern="quant_conv.conv3d.weight",
                from_pattern=["quant_conv.weight"],
                transform=WeightTransforms.transpose_conv3d_weight,
            ),
            WeightTarget(
                to_pattern="quant_conv.conv3d.bias",
                from_pattern=["quant_conv.bias"],
            ),
        ]

    @staticmethod
    def get_text_encoder_mapping() -> List[WeightTarget]:
        return [
            WeightTarget(
                to_pattern="encoder.embed_tokens.weight",
                from_pattern=["model.embed_tokens.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.norm.weight",
                from_pattern=["model.norm.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.input_layernorm.weight",
                from_pattern=["model.layers.{layer}.input_layernorm.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.post_attention_layernorm.weight",
                from_pattern=["model.layers.{layer}.post_attention_layernorm.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.self_attn.q_proj.weight",
                from_pattern=["model.layers.{layer}.self_attn.q_proj.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.self_attn.q_proj.bias",
                from_pattern=["model.layers.{layer}.self_attn.q_proj.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.self_attn.k_proj.weight",
                from_pattern=["model.layers.{layer}.self_attn.k_proj.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.self_attn.k_proj.bias",
                from_pattern=["model.layers.{layer}.self_attn.k_proj.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.self_attn.v_proj.weight",
                from_pattern=["model.layers.{layer}.self_attn.v_proj.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.self_attn.v_proj.bias",
                from_pattern=["model.layers.{layer}.self_attn.v_proj.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.self_attn.o_proj.weight",
                from_pattern=["model.layers.{layer}.self_attn.o_proj.weight"],
            ),
            # MLP
            WeightTarget(
                to_pattern="encoder.layers.{layer}.mlp.gate_proj.weight",
                from_pattern=["model.layers.{layer}.mlp.gate_proj.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.mlp.up_proj.weight",
                from_pattern=["model.layers.{layer}.mlp.up_proj.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.layers.{layer}.mlp.down_proj.weight",
                from_pattern=["model.layers.{layer}.mlp.down_proj.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.visual.patch_embed.proj.weight",
                from_pattern=["visual.patch_embed.proj.weight"],
                transform=WeightTransforms.transpose_patch_embed,
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.attn.qkv.weight",
                from_pattern=["visual.blocks.{block}.attn.qkv.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.attn.qkv.bias",
                from_pattern=["visual.blocks.{block}.attn.qkv.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.attn.proj.weight",
                from_pattern=["visual.blocks.{block}.attn.proj.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.attn.proj.bias",
                from_pattern=["visual.blocks.{block}.attn.proj.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.mlp.gate_proj.weight",
                from_pattern=["visual.blocks.{block}.mlp.gate_proj.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.mlp.gate_proj.bias",
                from_pattern=["visual.blocks.{block}.mlp.gate_proj.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.mlp.up_proj.weight",
                from_pattern=["visual.blocks.{block}.mlp.up_proj.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.mlp.up_proj.bias",
                from_pattern=["visual.blocks.{block}.mlp.up_proj.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.mlp.down_proj.weight",
                from_pattern=["visual.blocks.{block}.mlp.down_proj.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.mlp.down_proj.bias",
                from_pattern=["visual.blocks.{block}.mlp.down_proj.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.norm1.weight",
                from_pattern=["visual.blocks.{block}.norm1.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.blocks.{block}.norm2.weight",
                from_pattern=["visual.blocks.{block}.norm2.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.merger.ln_q.weight",
                from_pattern=["visual.merger.ln_q.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.merger.mlp_0.weight",
                from_pattern=["visual.merger.mlp.0.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.merger.mlp_0.bias",
                from_pattern=["visual.merger.mlp.0.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.merger.mlp_1.weight",
                from_pattern=["visual.merger.mlp.2.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.visual.merger.mlp_1.bias",
                from_pattern=["visual.merger.mlp.2.bias"],
                required=False,
            ),
        ]

    @staticmethod
    def get_mapping() -> List[WeightTarget]:
        return QwenWeightMapping.get_transformer_mapping()



from typing import List

from mflux.models.common.weights.mapping.weight_mapping import WeightMapping, WeightTarget
from mflux.models.common.weights.mapping.weight_transforms import WeightTransforms


class FluxWeightMapping(WeightMapping):
    @staticmethod
    def get_transformer_mapping() -> List[WeightTarget]:
        return [
            WeightTarget(
                to_pattern="x_embedder.weight",
                from_pattern=["x_embedder.weight"],
            ),
            WeightTarget(
                to_pattern="x_embedder.bias",
                from_pattern=["x_embedder.bias"],
            ),
            WeightTarget(
                to_pattern="context_embedder.weight",
                from_pattern=["context_embedder.weight"],
            ),
            WeightTarget(
                to_pattern="context_embedder.bias",
                from_pattern=["context_embedder.bias"],
            ),
            WeightTarget(
                to_pattern="proj_out.weight",
                from_pattern=["proj_out.weight"],
            ),
            WeightTarget(
                to_pattern="proj_out.bias",
                from_pattern=["proj_out.bias"],
            ),
            # TimeTextEmbed
            WeightTarget(
                to_pattern="time_text_embed.timestep_embedder.linear_1.weight",
                from_pattern=["time_text_embed.timestep_embedder.linear_1.weight"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.timestep_embedder.linear_1.bias",
                from_pattern=["time_text_embed.timestep_embedder.linear_1.bias"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.timestep_embedder.linear_2.weight",
                from_pattern=["time_text_embed.timestep_embedder.linear_2.weight"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.timestep_embedder.linear_2.bias",
                from_pattern=["time_text_embed.timestep_embedder.linear_2.bias"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.text_embedder.linear_1.weight",
                from_pattern=["time_text_embed.text_embedder.linear_1.weight"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.text_embedder.linear_1.bias",
                from_pattern=["time_text_embed.text_embedder.linear_1.bias"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.text_embedder.linear_2.weight",
                from_pattern=["time_text_embed.text_embedder.linear_2.weight"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.text_embedder.linear_2.bias",
                from_pattern=["time_text_embed.text_embedder.linear_2.bias"],
            ),
            WeightTarget(
                to_pattern="time_text_embed.guidance_embedder.linear_1.weight",
                from_pattern=["time_text_embed.guidance_embedder.linear_1.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="time_text_embed.guidance_embedder.linear_1.bias",
                from_pattern=["time_text_embed.guidance_embedder.linear_1.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="time_text_embed.guidance_embedder.linear_2.weight",
                from_pattern=["time_text_embed.guidance_embedder.linear_2.weight"],
                required=False,
            ),
            WeightTarget(
                to_pattern="time_text_embed.guidance_embedder.linear_2.bias",
                from_pattern=["time_text_embed.guidance_embedder.linear_2.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="norm_out.linear.weight",
                from_pattern=["norm_out.linear.weight"],
            ),
            WeightTarget(
                to_pattern="norm_out.linear.bias",
                from_pattern=["norm_out.linear.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.norm1.linear.weight",
                from_pattern=["transformer_blocks.{block}.norm1.linear.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.norm1.linear.bias",
                from_pattern=["transformer_blocks.{block}.norm1.linear.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.norm1_context.linear.weight",
                from_pattern=["transformer_blocks.{block}.norm1_context.linear.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.norm1_context.linear.bias",
                from_pattern=["transformer_blocks.{block}.norm1_context.linear.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_q.weight",
                from_pattern=["transformer_blocks.{block}.attn.to_q.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_q.bias",
                from_pattern=["transformer_blocks.{block}.attn.to_q.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_k.weight",
                from_pattern=["transformer_blocks.{block}.attn.to_k.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_k.bias",
                from_pattern=["transformer_blocks.{block}.attn.to_k.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_v.weight",
                from_pattern=["transformer_blocks.{block}.attn.to_v.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_v.bias",
                from_pattern=["transformer_blocks.{block}.attn.to_v.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_out.0.weight",
                from_pattern=["transformer_blocks.{block}.attn.to_out.0.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_out.0.bias",
                from_pattern=["transformer_blocks.{block}.attn.to_out.0.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_q_proj.weight",
                from_pattern=["transformer_blocks.{block}.attn.add_q_proj.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_q_proj.bias",
                from_pattern=["transformer_blocks.{block}.attn.add_q_proj.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_k_proj.weight",
                from_pattern=["transformer_blocks.{block}.attn.add_k_proj.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_k_proj.bias",
                from_pattern=["transformer_blocks.{block}.attn.add_k_proj.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_v_proj.weight",
                from_pattern=["transformer_blocks.{block}.attn.add_v_proj.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.add_v_proj.bias",
                from_pattern=["transformer_blocks.{block}.attn.add_v_proj.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_add_out.weight",
                from_pattern=["transformer_blocks.{block}.attn.to_add_out.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.to_add_out.bias",
                from_pattern=["transformer_blocks.{block}.attn.to_add_out.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.norm_q.weight",
                from_pattern=["transformer_blocks.{block}.attn.norm_q.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.norm_k.weight",
                from_pattern=["transformer_blocks.{block}.attn.norm_k.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.norm_added_q.weight",
                from_pattern=["transformer_blocks.{block}.attn.norm_added_q.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.attn.norm_added_k.weight",
                from_pattern=["transformer_blocks.{block}.attn.norm_added_k.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.ff.linear1.weight",
                from_pattern=["transformer_blocks.{block}.ff.net.0.proj.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.ff.linear1.bias",
                from_pattern=["transformer_blocks.{block}.ff.net.0.proj.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.ff.linear2.weight",
                from_pattern=["transformer_blocks.{block}.ff.net.2.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.ff.linear2.bias",
                from_pattern=["transformer_blocks.{block}.ff.net.2.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.ff_context.linear1.weight",
                from_pattern=["transformer_blocks.{block}.ff_context.net.0.proj.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.ff_context.linear1.bias",
                from_pattern=["transformer_blocks.{block}.ff_context.net.0.proj.bias"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.ff_context.linear2.weight",
                from_pattern=["transformer_blocks.{block}.ff_context.net.2.weight"],
            ),
            WeightTarget(
                to_pattern="transformer_blocks.{block}.ff_context.linear2.bias",
                from_pattern=["transformer_blocks.{block}.ff_context.net.2.bias"],
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.norm.linear.weight",
                from_pattern=["single_transformer_blocks.{block}.norm.linear.weight"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.norm.linear.bias",
                from_pattern=["single_transformer_blocks.{block}.norm.linear.bias"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.attn.to_q.weight",
                from_pattern=["single_transformer_blocks.{block}.attn.to_q.weight"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.attn.to_q.bias",
                from_pattern=["single_transformer_blocks.{block}.attn.to_q.bias"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.attn.to_k.weight",
                from_pattern=["single_transformer_blocks.{block}.attn.to_k.weight"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.attn.to_k.bias",
                from_pattern=["single_transformer_blocks.{block}.attn.to_k.bias"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.attn.to_v.weight",
                from_pattern=["single_transformer_blocks.{block}.attn.to_v.weight"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.attn.to_v.bias",
                from_pattern=["single_transformer_blocks.{block}.attn.to_v.bias"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.attn.norm_q.weight",
                from_pattern=["single_transformer_blocks.{block}.attn.norm_q.weight"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.attn.norm_k.weight",
                from_pattern=["single_transformer_blocks.{block}.attn.norm_k.weight"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.proj_mlp.weight",
                from_pattern=["single_transformer_blocks.{block}.proj_mlp.weight"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.proj_mlp.bias",
                from_pattern=["single_transformer_blocks.{block}.proj_mlp.bias"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.proj_out.weight",
                from_pattern=["single_transformer_blocks.{block}.proj_out.weight"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="single_transformer_blocks.{block}.proj_out.bias",
                from_pattern=["single_transformer_blocks.{block}.proj_out.bias"],
                max_blocks=38,
            ),
        ]

    @staticmethod
    def get_controlnet_transformer_mapping() -> List[WeightTarget]:
        return FluxWeightMapping.get_transformer_mapping() + [
            WeightTarget(
                to_pattern="controlnet_x_embedder.weight",
                from_pattern=["controlnet_x_embedder.weight"],
            ),
            WeightTarget(
                to_pattern="controlnet_x_embedder.bias",
                from_pattern=["controlnet_x_embedder.bias"],
            ),
            WeightTarget(
                to_pattern="controlnet_blocks.{block}.weight",
                from_pattern=["controlnet_blocks.{block}.weight"],
                max_blocks=19,
            ),
            WeightTarget(
                to_pattern="controlnet_blocks.{block}.bias",
                from_pattern=["controlnet_blocks.{block}.bias"],
                max_blocks=19,
            ),
            WeightTarget(
                to_pattern="controlnet_single_blocks.{block}.weight",
                from_pattern=["controlnet_single_blocks.{block}.weight"],
                max_blocks=38,
            ),
            WeightTarget(
                to_pattern="controlnet_single_blocks.{block}.bias",
                from_pattern=["controlnet_single_blocks.{block}.bias"],
                max_blocks=38,
            ),
        ]

    @staticmethod
    def get_vae_mapping() -> List[WeightTarget]:
        return [
            WeightTarget(
                to_pattern="decoder.conv_in.conv2d.weight",
                from_pattern=["decoder.conv_in.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.conv_in.conv2d.bias",
                from_pattern=["decoder.conv_in.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.conv_out.conv2d.weight",
                from_pattern=["decoder.conv_out.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.conv_out.conv2d.bias",
                from_pattern=["decoder.conv_out.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.conv_norm_out.norm.weight",
                from_pattern=["decoder.conv_norm_out.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.conv_norm_out.norm.bias",
                from_pattern=["decoder.conv_norm_out.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.conv_in.conv2d.weight",
                from_pattern=["encoder.conv_in.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.conv_in.conv2d.bias",
                from_pattern=["encoder.conv_in.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.conv_out.conv2d.weight",
                from_pattern=["encoder.conv_out.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.conv_out.conv2d.bias",
                from_pattern=["encoder.conv_out.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.conv_norm_out.norm.weight",
                from_pattern=["encoder.conv_norm_out.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.conv_norm_out.norm.bias",
                from_pattern=["encoder.conv_norm_out.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.norm1.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.norm1.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.norm1.bias",
                from_pattern=["decoder.mid_block.resnets.{i}.norm1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv1.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.conv1.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv1.bias",
                from_pattern=["decoder.mid_block.resnets.{i}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.norm2.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.norm2.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.norm2.bias",
                from_pattern=["decoder.mid_block.resnets.{i}.norm2.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv2.weight",
                from_pattern=["decoder.mid_block.resnets.{i}.conv2.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.resnets.{i}.conv2.bias",
                from_pattern=["decoder.mid_block.resnets.{i}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.norm1.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.norm1.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.norm1.bias",
                from_pattern=["encoder.mid_block.resnets.{i}.norm1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv1.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.conv1.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv1.bias",
                from_pattern=["encoder.mid_block.resnets.{i}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.norm2.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.norm2.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.norm2.bias",
                from_pattern=["encoder.mid_block.resnets.{i}.norm2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv2.weight",
                from_pattern=["encoder.mid_block.resnets.{i}.conv2.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.resnets.{i}.conv2.bias",
                from_pattern=["encoder.mid_block.resnets.{i}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.group_norm.weight",
                from_pattern=["decoder.mid_block.attentions.0.group_norm.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.group_norm.bias",
                from_pattern=["decoder.mid_block.attentions.0.group_norm.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_q.weight",
                from_pattern=["decoder.mid_block.attentions.0.to_q.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_q.bias",
                from_pattern=["decoder.mid_block.attentions.0.to_q.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_k.weight",
                from_pattern=["decoder.mid_block.attentions.0.to_k.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_k.bias",
                from_pattern=["decoder.mid_block.attentions.0.to_k.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_v.weight",
                from_pattern=["decoder.mid_block.attentions.0.to_v.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_v.bias",
                from_pattern=["decoder.mid_block.attentions.0.to_v.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_out.0.weight",
                from_pattern=["decoder.mid_block.attentions.0.to_out.0.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.mid_block.attentions.0.to_out.0.bias",
                from_pattern=["decoder.mid_block.attentions.0.to_out.0.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.group_norm.weight",
                from_pattern=["encoder.mid_block.attentions.0.group_norm.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.group_norm.bias",
                from_pattern=["encoder.mid_block.attentions.0.group_norm.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_q.weight",
                from_pattern=["encoder.mid_block.attentions.0.to_q.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_q.bias",
                from_pattern=["encoder.mid_block.attentions.0.to_q.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_k.weight",
                from_pattern=["encoder.mid_block.attentions.0.to_k.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_k.bias",
                from_pattern=["encoder.mid_block.attentions.0.to_k.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_v.weight",
                from_pattern=["encoder.mid_block.attentions.0.to_v.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_v.bias",
                from_pattern=["encoder.mid_block.attentions.0.to_v.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_out.0.weight",
                from_pattern=["encoder.mid_block.attentions.0.to_out.0.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.mid_block.attentions.0.to_out.0.bias",
                from_pattern=["encoder.mid_block.attentions.0.to_out.0.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.norm1.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.norm1.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.norm1.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.norm1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv1.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv1.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv1.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.norm2.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.norm2.weight"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.norm2.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.norm2.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv2.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv2.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv2.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv_shortcut.weight",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv_shortcut.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.resnets.{res}.conv_shortcut.bias",
                from_pattern=["decoder.up_blocks.{block}.resnets.{res}.conv_shortcut.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.upsamplers.0.conv.weight",
                from_pattern=["decoder.up_blocks.{block}.upsamplers.0.conv.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="decoder.up_blocks.{block}.upsamplers.0.conv.bias",
                from_pattern=["decoder.up_blocks.{block}.upsamplers.0.conv.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.norm1.weight",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.norm1.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.norm1.bias",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.norm1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv1.weight",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv1.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv1.bias",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv1.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.norm2.weight",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.norm2.weight"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.norm2.bias",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.norm2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv2.weight",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv2.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv2.bias",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv2.bias"],
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv_shortcut.weight",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv_shortcut.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.resnets.{res}.conv_shortcut.bias",
                from_pattern=["encoder.down_blocks.{block}.resnets.{res}.conv_shortcut.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.downsamplers.0.conv.weight",
                from_pattern=["encoder.down_blocks.{block}.downsamplers.0.conv.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
                required=False,
            ),
            WeightTarget(
                to_pattern="encoder.down_blocks.{block}.downsamplers.0.conv.bias",
                from_pattern=["encoder.down_blocks.{block}.downsamplers.0.conv.bias"],
                required=False,
            ),
            WeightTarget(
                to_pattern="quant_conv.weight",
                from_pattern=["quant_conv.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="quant_conv.bias",
                from_pattern=["quant_conv.bias"],
            ),
            WeightTarget(
                to_pattern="post_quant_conv.weight",
                from_pattern=["post_quant_conv.weight"],
                transform=WeightTransforms.transpose_conv2d_weight,
            ),
            WeightTarget(
                to_pattern="post_quant_conv.bias",
                from_pattern=["post_quant_conv.bias"],
            ),
        ]

    @staticmethod
    def get_t5_encoder_mapping() -> List[WeightTarget]:
        return [
            WeightTarget(
                to_pattern="shared.weight",
                from_pattern=["shared.weight"],
            ),
            WeightTarget(
                to_pattern="final_layer_norm.weight",
                from_pattern=["encoder.final_layer_norm.weight"],
            ),
            WeightTarget(
                to_pattern="t5_blocks.{block}.attention.layer_norm.weight",
                from_pattern=["encoder.block.{block}.layer.0.layer_norm.weight"],
            ),
            WeightTarget(
                to_pattern="t5_blocks.{block}.attention.SelfAttention.q.weight",
                from_pattern=["encoder.block.{block}.layer.0.SelfAttention.q.weight"],
            ),
            WeightTarget(
                to_pattern="t5_blocks.{block}.attention.SelfAttention.k.weight",
                from_pattern=["encoder.block.{block}.layer.0.SelfAttention.k.weight"],
            ),
            WeightTarget(
                to_pattern="t5_blocks.{block}.attention.SelfAttention.v.weight",
                from_pattern=["encoder.block.{block}.layer.0.SelfAttention.v.weight"],
            ),
            WeightTarget(
                to_pattern="t5_blocks.{block}.attention.SelfAttention.o.weight",
                from_pattern=["encoder.block.{block}.layer.0.SelfAttention.o.weight"],
            ),
            WeightTarget(
                to_pattern="t5_blocks.{block}.attention.SelfAttention.relative_attention_bias.weight",
                from_pattern=["encoder.block.0.layer.0.SelfAttention.relative_attention_bias.weight"],
                max_blocks=24,
            ),
            WeightTarget(
                to_pattern="t5_blocks.{block}.ff.layer_norm.weight",
                from_pattern=["encoder.block.{block}.layer.1.layer_norm.weight"],
            ),
            WeightTarget(
                to_pattern="t5_blocks.{block}.ff.DenseReluDense.wi_0.weight",
                from_pattern=["encoder.block.{block}.layer.1.DenseReluDense.wi_0.weight"],
            ),
            WeightTarget(
                to_pattern="t5_blocks.{block}.ff.DenseReluDense.wi_1.weight",
                from_pattern=["encoder.block.{block}.layer.1.DenseReluDense.wi_1.weight"],
            ),
            WeightTarget(
                to_pattern="t5_blocks.{block}.ff.DenseReluDense.wo.weight",
                from_pattern=["encoder.block.{block}.layer.1.DenseReluDense.wo.weight"],
            ),
        ]

    @staticmethod
    def get_clip_encoder_mapping() -> List[WeightTarget]:
        return [
            WeightTarget(
                to_pattern="text_model.embeddings.token_embedding.weight",
                from_pattern=["text_model.embeddings.token_embedding.weight"],
            ),
            WeightTarget(
                to_pattern="text_model.embeddings.position_embedding.weight",
                from_pattern=["text_model.embeddings.position_embedding.weight"],
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.self_attn.q_proj.weight",
                from_pattern=["text_model.encoder.layers.{block}.self_attn.q_proj.weight"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.self_attn.q_proj.bias",
                from_pattern=["text_model.encoder.layers.{block}.self_attn.q_proj.bias"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.self_attn.k_proj.weight",
                from_pattern=["text_model.encoder.layers.{block}.self_attn.k_proj.weight"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.self_attn.k_proj.bias",
                from_pattern=["text_model.encoder.layers.{block}.self_attn.k_proj.bias"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.self_attn.v_proj.weight",
                from_pattern=["text_model.encoder.layers.{block}.self_attn.v_proj.weight"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.self_attn.v_proj.bias",
                from_pattern=["text_model.encoder.layers.{block}.self_attn.v_proj.bias"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.self_attn.out_proj.weight",
                from_pattern=["text_model.encoder.layers.{block}.self_attn.out_proj.weight"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.self_attn.out_proj.bias",
                from_pattern=["text_model.encoder.layers.{block}.self_attn.out_proj.bias"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.layer_norm1.weight",
                from_pattern=["text_model.encoder.layers.{block}.layer_norm1.weight"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.layer_norm1.bias",
                from_pattern=["text_model.encoder.layers.{block}.layer_norm1.bias"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.mlp.fc1.weight",
                from_pattern=["text_model.encoder.layers.{block}.mlp.fc1.weight"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.mlp.fc1.bias",
                from_pattern=["text_model.encoder.layers.{block}.mlp.fc1.bias"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.mlp.fc2.weight",
                from_pattern=["text_model.encoder.layers.{block}.mlp.fc2.weight"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.mlp.fc2.bias",
                from_pattern=["text_model.encoder.layers.{block}.mlp.fc2.bias"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.layer_norm2.weight",
                from_pattern=["text_model.encoder.layers.{block}.layer_norm2.weight"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.encoder.layers.{block}.layer_norm2.bias",
                from_pattern=["text_model.encoder.layers.{block}.layer_norm2.bias"],
                max_blocks=12,
            ),
            WeightTarget(
                to_pattern="text_model.final_layer_norm.weight",
                from_pattern=["text_model.final_layer_norm.weight"],
            ),
            WeightTarget(
                to_pattern="text_model.final_layer_norm.bias",
                from_pattern=["text_model.final_layer_norm.bias"],
            ),
        ]



import mlx.core as mx
from mlx import nn

from mflux.models.flux.model.flux_text_encoder.clip_encoder.clip_encoder_layer import CLIPEncoderLayer


class EncoderCLIP(nn.Module):
    def __init__(self, num_encoder_layers: int):
        super().__init__()
        self.layers = [CLIPEncoderLayer(i) for i in range(num_encoder_layers)]

    def __call__(self, tokens: mx.array, causal_attention_mask: mx.array) -> mx.array:
        hidden_states = tokens
        for encoder_layer in self.layers:
            hidden_states = encoder_layer(hidden_states, causal_attention_mask)
        return hidden_states



from typing import List

import mlx.nn as nn

from mflux.models.common.config.model_config import ModelConfig
from mflux.models.common.tokenizer import LanguageTokenizer
from mflux.models.common.weights.loading.weight_definition import ComponentDefinition, TokenizerDefinition
from mflux.models.common.weights.mapping.weight_transforms import WeightTransforms
from mflux.models.flux.weights.flux_weight_mapping import FluxWeightMapping


class FluxWeightDefinition:
    @staticmethod
    def get_components() -> List[ComponentDefinition]:
        return [
            ComponentDefinition(
                name="vae",
                hf_subdir="vae",
                precision=ModelConfig.precision,
                mapping_getter=FluxWeightMapping.get_vae_mapping,
            ),
            ComponentDefinition(
                name="transformer",
                hf_subdir="transformer",
                precision=ModelConfig.precision,
                mapping_getter=FluxWeightMapping.get_transformer_mapping,
            ),
            ComponentDefinition(
                name="t5_encoder",
                hf_subdir="text_encoder_2",
                model_attr="t5_text_encoder",
                num_blocks=24,
                precision=ModelConfig.precision,
                mapping_getter=FluxWeightMapping.get_t5_encoder_mapping,
            ),
            ComponentDefinition(
                name="clip_encoder",
                hf_subdir="text_encoder",
                model_attr="clip_text_encoder",
                precision=ModelConfig.precision,
                mapping_getter=FluxWeightMapping.get_clip_encoder_mapping,
            ),
        ]

    @staticmethod
    def get_tokenizers() -> List[TokenizerDefinition]:
        return [
            TokenizerDefinition(
                name="clip",
                hf_subdir="tokenizer",
                tokenizer_class="CLIPTokenizer",
                encoder_class=LanguageTokenizer,
                max_length=77,
                download_patterns=["tokenizer/**"],
            ),
            TokenizerDefinition(
                name="t5",
                hf_subdir="tokenizer_2",
                tokenizer_class="T5Tokenizer",
                encoder_class=LanguageTokenizer,
                max_length=256,  # Will be overridden by model_config.max_sequence_length
                download_patterns=["tokenizer_2/**"],
            ),
        ]

    @staticmethod
    def get_download_patterns() -> List[str]:
        return [
            "text_encoder/*.safetensors",
            "text_encoder/*.json",
            "text_encoder_2/*.safetensors",
            "text_encoder_2/*.json",
            "transformer/*.safetensors",
            "transformer/*.json",
            "vae/*.safetensors",
            "vae/*.json",
        ]

    @staticmethod
    def quantization_predicate(path: str, module) -> bool:
        return hasattr(module, "to_quantized")


class FluxControlnetWeightDefinition:
    @staticmethod
    def get_controlnet_component() -> ComponentDefinition:
        return ComponentDefinition(
            name="transformer_controlnet",
            hf_subdir="",
            loading_mode="single",
            precision=ModelConfig.precision,
            mapping_getter=FluxWeightMapping.get_controlnet_transformer_mapping,
        )

    @staticmethod
    def get_components() -> List[ComponentDefinition]:
        return FluxWeightDefinition.get_components() + [
            ComponentDefinition(
                name="transformer_controlnet",
                hf_subdir="transformer_controlnet",
                precision=ModelConfig.precision,
                mapping_getter=FluxWeightMapping.get_transformer_mapping,
            ),
        ]

    @staticmethod
    def get_tokenizers() -> List[TokenizerDefinition]:
        return FluxWeightDefinition.get_tokenizers()

    @staticmethod
    def get_download_patterns() -> List[str]:
        return FluxWeightDefinition.get_download_patterns()

    @staticmethod
    def quantization_predicate(path: str, module) -> bool:
        return FluxWeightDefinition.quantization_predicate(path, module)


class FluxReduxWeightDefinition:
    @staticmethod
    def get_components() -> List[ComponentDefinition]:
        return [
            ComponentDefinition(
                name="siglip",
                hf_subdir="image_encoder",
                mapping_getter=None,
                model_attr="image_encoder",
                precision=ModelConfig.precision,
                bulk_transform=WeightTransforms.transpose_conv2d_weight,
                weight_subkey="vision_model",
            ),
            ComponentDefinition(
                name="redux_encoder",
                hf_subdir="image_embedder",
                mapping_getter=None,
                model_attr="image_embedder",
                precision=ModelConfig.precision,
            ),
        ]

    @staticmethod
    def get_tokenizers() -> List[TokenizerDefinition]:
        return []

    @staticmethod
    def get_download_patterns() -> List[str]:
        return [
            "image_encoder/*.safetensors",
            "image_encoder/config.json",
            "image_embedder/*.safetensors",
            "image_embedder/config.json",
        ]

    @staticmethod
    def quantization_predicate(path: str, module) -> bool:
        if isinstance(module, nn.Conv2d):
            return False
        if hasattr(module, "weight") and hasattr(module.weight, "shape"):
            if module.weight.shape == (1152, 4304):
                return False
            if module.weight.shape[-1] % 64 != 0:
                return False
        return hasattr(module, "to_quantized")



import mlx.core as mx
from mlx import nn


class ConvOut(nn.Module):
    def __init__(self, in_channels: int = 128, out_channels: int = 3):
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
        )

    def __call__(self, input_array: mx.array) -> mx.array:
        input_array = mx.transpose(input_array, (0, 2, 3, 1))
        hidden_states = self.conv(input_array)
        return mx.transpose(hidden_states, (0, 3, 1, 2))



import mlx.core as mx
from mlx import nn

from mflux.models.seedvr2.model.seedvr2_vae.encoder.downsample_3d import Downsample3D
from mflux.models.seedvr2.model.seedvr2_vae.encoder.resnet_block_3d import ResnetBlock3D


class DownBlock3D(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_layers: int = 2,
        add_downsample: bool = True,
        temporal_down: bool = False,
    ):
        super().__init__()
        self.resnets = []
        for i in range(num_layers):
            in_ch = in_channels if i == 0 else out_channels
            self.resnets.append(ResnetBlock3D(in_channels=in_ch, out_channels=out_channels))

        self.downsamplers = []
        if add_downsample:
            self.downsamplers.append(Downsample3D(channels=out_channels, spatial_only=not temporal_down))

    def __call__(self, x: mx.array) -> mx.array:
        for resnet in self.resnets:
            x = resnet(x)
        for downsampler in self.downsamplers:
            x = downsampler(x)
        return x



from typing import TYPE_CHECKING

import mlx.core as mx

if TYPE_CHECKING:
    from mflux.models.common.config.config import Config

from mflux.models.common.schedulers.base_scheduler import BaseScheduler


class LinearScheduler(BaseScheduler):
    def __init__(self, config: "Config"):
        self.config = config
        self._sigmas = self._get_sigmas()
        self._timesteps = self._get_timesteps()

    @property
    def sigmas(self) -> mx.array:
        return self._sigmas

    @property
    def timesteps(self) -> mx.array:
        return self._timesteps

    def _get_sigmas(self) -> mx.array:
        model_config = self.config.model_config
        sigmas = mx.linspace(
            1.0,
            1.0 / self.config.num_inference_steps,
            self.config.num_inference_steps,
        )
        sigmas = mx.array(sigmas).astype(mx.float32)
        sigmas = mx.concatenate([sigmas, mx.zeros(1)])
        if model_config.requires_sigma_shift:
            y1 = 0.5
            x1 = 256
            m = (1.15 - y1) / (4096 - x1)
            b = y1 - m * x1
            mu = m * self.config.width * self.config.height / 256 + b
            mu = mx.array(mu)
            shifted_sigmas = mx.exp(mu) / (mx.exp(mu) + (1 / sigmas - 1))
            shifted_sigmas[-1] = 0
            return shifted_sigmas
        else:
            return sigmas

    def _get_timesteps(self) -> mx.array:
        num_steps = self.config.num_inference_steps
        timesteps = mx.arange(num_steps, dtype=mx.float32)

        return timesteps

    def step(self, noise: mx.array, timestep: int, latents: mx.array, **kwargs) -> mx.array:
        dt = self._sigmas[timestep + 1] - self._sigmas[timestep]
        return latents + noise * dt



import mlx.core as mx
from mlx import nn


class FiboAdaLayerNormZero(nn.Module):
    def __init__(self, embedding_dim: int, eps: float = 1e-6):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.eps = eps
        self.linear = nn.Linear(embedding_dim, 6 * embedding_dim)

    def __call__(self, hidden_states: mx.array, text_embeddings: mx.array):
        emb = self.linear(nn.silu(text_embeddings))
        chunk = self.embedding_dim
        shift_msa = emb[:, 0 * chunk : 1 * chunk]
        scale_msa = emb[:, 1 * chunk : 2 * chunk]
        gate_msa = emb[:, 2 * chunk : 3 * chunk]
        shift_mlp = emb[:, 3 * chunk : 4 * chunk]
        scale_mlp = emb[:, 4 * chunk : 5 * chunk]
        gate_mlp = emb[:, 5 * chunk : 6 * chunk]
        norm_hidden_states = FiboAdaLayerNormZero._layer_norm(self.eps, hidden_states)
        hidden_states = norm_hidden_states * (1 + scale_msa[:, None, :]) + shift_msa[:, None, :]
        return hidden_states, gate_msa, shift_mlp, scale_mlp, gate_mlp

    @staticmethod
    def _layer_norm(eps, x: mx.array) -> mx.array:
        x_f32 = x.astype(mx.float32)
        mean = mx.mean(x_f32, axis=-1, keepdims=True)
        var = mx.mean((x_f32 - mean) ** 2, axis=-1, keepdims=True)
        y = (x_f32 - mean) / mx.sqrt(var + eps)
        return y.astype(x.dtype)
